from IPython.core.magic import register_line_magic
import subprocess
from imports.analyze import (
    _get_dists,
    _build_top_module_to_dep_map,
    get_imported_modules,
)

from typing import Dict


def _dict_diff(d1, d2) -> Dict[str, any]:
    keys1 = set(d1.keys())
    keys2 = set(d2.keys())

    added = keys2 - keys1
    removed = keys1 - keys2
    modified = {k: (d1[k], d2[k]) for k in keys1 & keys2 if d1[k] != d2[k]}

    if not (added or removed or modified):
        return {}

    return {
        "added": {k: d2[k] for k in added},
        "removed": {k: d1[k] for k in removed},
        "modified": modified,
    }


@register_line_magic
def smart_pip(line):
    dists_snapshot = _get_dists()
    ret = subprocess.run(["pip"] + line.split(" "))
    if ret.returncode:
        print(f"Failed to run pip {line}. stdout: {ret.stdout} | stderr: {ret.stderr}")

    dists_delta = _dict_diff(dists_snapshot, _get_dists())

    if "added" in dists_delta:
        affected_modules = _build_top_module_to_dep_map(dists_delta["added"]).keys()
        imported_modules = set(get_imported_modules())
        modules_to_reimport = imported_modules & affected_modules
        if modules_to_reimport:
            return f'You should restart Python because the underlying files are updated for these dependencies: {','.join(modules_to_reimport)}'
    return ""



import sys
import os

from typing import List, Dict, Tuple

__all__ = ["get_imported_modules", "get_imported_depdencies"]


def get_imported_modules() -> List[str]:
    return list(sys.modules.keys())


def get_imported_depdencies() -> Dict[str, any]:
    result = {}
    top_module_to_dep = _build_top_module_to_dep_map()
    for mod in get_imported_modules():
        if mod in top_module_to_dep:
            dep, version = top_module_to_dep[mod]
            result.update({mod: (dep, version)})
    return result


def _get_dists() -> Dict[str, str]:
    site_package_path = list(filter(lambda p: "site-packages" in p, sys.path))[0]
    dist_info_folders = list(
        filter(lambda folder: "dist-info" in folder, os.listdir(site_package_path))
    )
    return dict(
        map(
            lambda folder: (
                folder.removesuffix(".dist-info"),
                f"{site_package_path}/{folder}",
            ),
            dist_info_folders,
        )
    )


def _build_top_module_to_dep_map(
    dists: Dict[str, str] = None,
) -> Dict[str, Tuple[str, str]]:
    top_module_to_dep = {}
    for dep_version, path in (dists or _get_dists()).items():
        dep, version = dep_version.split("-")
        with open(f"{path}/RECORD") as f:
            for line in f.readlines():
                source = line.strip().split(",")[0]
                if source.endswith(".py"):
                    top_module = source.split("/")[0]
                    top_module_to_dep.update({top_module: (dep, version)})
    return top_module_to_dep
